"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from jax.core import ShapedArray
from jaxlib.xla_extension import ArrayImpl
from jax.extend.core import ClosedJaxpr

from qrisp.circuit import Qubit, QuantumCircuit, XGate
from qrisp.core.session_merging_tools import merge, merge_sessions, multi_session_merge
from qrisp.environments import QuantumEnvironment, ClControlEnvironment
from qrisp.misc import perm_lock, perm_unlock, bin_rep
from qrisp.jasp import check_for_tracing_mode, get_last_equation
from qrisp.core import mcx, p, rz, x

import numpy as np


class ControlEnvironment(QuantumEnvironment):
    """
    This class behaves similarly to ConditionEnvironment but instead of a function
    calculating a truth value, we supply a list of qubits.
    The environment's content is then controlled on these qubits.

    An alias for this QuantumEnvironment is "control".


    Parameters
    ----------
    ctrl_qubits : list[Qubit]
        A list of qubits on which to control the environment's content.
    ctrl_state : int/str, optional
        The computational basis state which is supposed to activate the environment.
        Can be supplied as a bitstring or integer. The default is "1111..".


    Examples
    --------

    We create a QuantumVariable and control on some of it's qubits
    using the control alias ::

        from qrisp import QuantumVariable, QuantumString, multi_measurement, control, h

        qv = QuantumVariable(3)
        q_str = QuantumString()

        qv[:] = "011"
        h(qv[0])

        with control(qv[:2], "11"):
            q_str += "hello world"


    >>> print(multi_measurement([qv, q_str]))
    {('011', 'aaaaaaaaaaa'): 0.5, ('111', 'hello world'): 0.5}


    """

    def __init__(self, ctrl_qubits, ctrl_state=-1, ctrl_method=None, invert=False):

        if isinstance(ctrl_state, int):
            if ctrl_state < 0:
                ctrl_state += 2 ** len(ctrl_qubits)

            self.ctrl_state = bin_rep(ctrl_state, len(ctrl_qubits))[::-1]
        else:
            self.ctrl_state = str(ctrl_state)

        if check_for_tracing_mode():

            QuantumEnvironment.__init__(self, list(ctrl_qubits))
            if not isinstance(ctrl_qubits, list):
                ctrl_qubits = [ctrl_qubits]

        else:

            if isinstance(ctrl_qubits, list):
                self.arg_qs = multi_session_merge([qb.qs() for qb in ctrl_qubits])
            else:
                self.arg_qs = ctrl_qubits.qs()
            self.arg_qs = merge(ctrl_qubits)

            self.ctrl_method = ctrl_method
            if isinstance(ctrl_qubits, Qubit):
                ctrl_qubits = [ctrl_qubits]

            self.ctrl_qubits = ctrl_qubits
            self.invert = invert

            self.manual_allocation_management = True

            # For more information on why this attribute is neccessary check the comment
            # on the line containing subcondition_truth_values = []
            self.sub_condition_envs = []

            QuantumEnvironment.__init__(self)

    # Method to enter the environment
    def __enter__(self):

        if check_for_tracing_mode():
            QuantumEnvironment.__enter__(self)
            return

        from qrisp.qtypes.quantum_bool import QuantumBool

        if len(self.ctrl_qubits) == 1:
            self.condition_truth_value = self.ctrl_qubits[0]
        else:
            self.qbool = QuantumBool(name="ctrl_env*", qs=self.arg_qs)
            self.condition_truth_value = self.qbool[0]

        QuantumEnvironment.__enter__(self)

        merge_sessions(self.env_qs, self.arg_qs)

        return self.condition_truth_value

    def __exit__(self, exception_type, exception_value, traceback):
        from qrisp.environments import (
            ConditionEnvironment,
            ControlEnvironment,
            InversionEnvironment,
        )

        if check_for_tracing_mode():
            QuantumEnvironment.__exit__(
                self, exception_type, exception_value, traceback
            )
            return
        self.parent_cond_env = None

        QuantumEnvironment.__exit__(self, exception_type, exception_value, traceback)

        from qrisp import ConjugationEnvironment

        # Determine the parent environment
        for env in self.env_qs.env_stack[::-1]:
            if isinstance(env, (ControlEnvironment, ConditionEnvironment)):
                self.parent_cond_env = env
                break
            if not isinstance(env, (InversionEnvironment, ConjugationEnvironment)):

                if not type(env) == QuantumEnvironment:
                    break

    def compile(self):
        from qrisp import QuantumBool
        from qrisp.environments import ConditionEnvironment, CustomControlOperation

        # Create the quantum variable where the condition truth value should be saved
        # Incase we have a parent environment we create two qubits because
        # we use the second qubit to compute the toffoli of this one and the parent
        # environments truth value in order to not have the environment operations
        # controlled on two qubits

        if len(self.env_data):
            # The first step we have to perform is calculating the truth value of the
            # environments quantum condition. For this we differentiate between
            # the case that this condition is embedded in another condition or not

            ctrl_qubits = list(self.ctrl_qubits)
            cond_compile_ctrl_state = self.ctrl_state

            if self.parent_cond_env is not None:

                # In the parent case we also need to make sure that the code is executed
                # if the parent environment is executed. A possible approach would be
                # to control the content on both, the parent and the chield truth value.
                # However, for each nesting level the gate count to generate
                # the controlled-controlled-controlled... version of the gates inside
                # the environment increases exponentially. Because of this we compute
                # the toffoli of the parent and child truth value
                # and controll the environment gates on this qubit.

                # Synthesize the condition of the environment
                # into the condition truth value qubit
                if len(ctrl_qubits) == 1:
                    from qrisp.misc import retarget_instructions

                    self.qbool = QuantumBool(name="ctrl_env*", qs=self.env_qs)
                    retarget_instructions(
                        self.env_data, [self.condition_truth_value], [self.qbool[0]]
                    )

                    if len(self.env_qs.data):
                        if isinstance(self.env_qs.data[-1], QuantumEnvironment):
                            env = self.env_qs.data.pop(-1)
                            env.compile()

                    self.condition_truth_value = self.qbool[0]

                ctrl_qubits.append(self.parent_cond_env.condition_truth_value)

                parent_ctrl_state = "1"

                if isinstance(self.parent_cond_env, ControlEnvironment):
                    if len(self.parent_cond_env.ctrl_qubits) == 1:
                        if self.parent_cond_env.ctrl_state == "0" and not hasattr(
                            self.parent_cond_env, "qbool"
                        ):
                            parent_ctrl_state = "0"

                    if self.parent_cond_env.invert:
                        if parent_ctrl_state == "0":
                            parent_ctrl_state = "1"
                        elif parent_ctrl_state == "1":
                            parent_ctrl_state = "0"

                cond_compile_ctrl_state = cond_compile_ctrl_state + parent_ctrl_state

            if len(ctrl_qubits) > 1:

                if len(ctrl_qubits) > 5:
                    method = "auto"
                else:
                    method = "gray_pt"

                mcx(
                    ctrl_qubits,
                    self.condition_truth_value,
                    ctrl_state=cond_compile_ctrl_state,
                    method=method,
                )

            perm_lock(ctrl_qubits)
            # unlock(self.condition_truth_value)

            # This list will contain the qubits holding the truth values of
            # conditional/control environments within this environment.
            # The instruction from the subcondition environments do not need to be
            # controlled, since their compile method compiles their condition
            # truth value based on the truth value of the parent environment.
            subcondition_truth_values = [
                env.condition_truth_value for env in self.sub_condition_envs
            ]
            inversion_tracker = 1

            # Now we need to recover the instructions from the data list
            # and perform their controlled version on the condition_truth_value qubit
            while self.env_data:
                instruction = self.env_data.pop(0)

                # If the instruction == conditional environment, compile the environment
                if isinstance(instruction, (ControlEnvironment, ConditionEnvironment)):
                    instruction.compile()

                    subcondition_truth_values = [
                        env.condition_truth_value for env in self.sub_condition_envs
                    ]
                    continue

                # If the instruction is a general environment, compile the instruction
                # and add the compilation result to the list of instructions
                # that need to be conditionally executed.
                elif issubclass(instruction.__class__, QuantumEnvironment):
                    temp_data_list = list(self.env_qs.data)
                    self.env_qs.clear_data()
                    instruction.compile()
                    self.env_data = list(self.env_qs.data) + self.env_data
                    self.env_qs.clear_data()
                    self.env_qs.data.extend(temp_data_list)

                    subcondition_truth_values = [
                        env.condition_truth_value for env in self.sub_condition_envs
                    ]
                    continue

                if (
                    instruction.op.name in ["qb_alloc", "qb_dealloc"]
                    and instruction.qubits[0] != self.condition_truth_value
                ) or instruction.op.name == "barrier":
                    self.env_qs.append(instruction)
                    continue

                if set(instruction.qubits).intersection(subcondition_truth_values):
                    # REWORK required: Condition environments evaluate their condition
                    # and (if within another control/condition environment) combine
                    # this QuantumBool (via pt toffoli) with the superior condition.
                    # This "combining" is done, so the condition evaluation doesn't
                    # have to be controlled. However, this is not properly realized
                    # in this method. The condition evaluation could probably work
                    # better using the custom control feature.
                    self.env_qs.append(instruction)
                    continue

                # Support for inversion of the condition without opening a new
                # environment
                # if set(instruction.qubits).issubset(self.user_exposed_qbool):
                if set(instruction.qubits).issubset(
                    [self.condition_truth_value]
                ) and not isinstance(instruction.op, CustomControlOperation):
                    if instruction.op.name == "x":
                        inversion_tracker *= -1
                        x(self.condition_truth_value)
                    elif instruction.op.name == "p":
                        p(instruction.op.params[0], self.condition_truth_value)
                    elif instruction.op.name == "rz":
                        rz(instruction.op.params[0], self.condition_truth_value)
                    elif instruction.op.name in ["qb_alloc", "qb_dealloc"]:
                        pass
                    else:
                        raise Exception(
                            f"Tried to perform invalid operation {instruction.op.name} "
                            "on condition truth value (allowed are x, p, rz)"
                        )
                    continue

                if len(ctrl_qubits) == 1:
                    ctrl_state = self.ctrl_state
                else:
                    ctrl_state = "1"

                if self.invert:

                    new_ctrl_state = ""
                    for c in ctrl_state:
                        if c == "1":
                            new_ctrl_state += "0"
                        else:
                            new_ctrl_state += "1"

                    ctrl_state = new_ctrl_state

                if self.condition_truth_value in instruction.qubits:
                    self.env_qs.append(
                        convert_to_custom_control(
                            instruction,
                            self.condition_truth_value,
                            invert_control=ctrl_state == "0",
                        )
                    )
                    continue

                else:
                    # Create controlled instruction
                    instruction.op = instruction.op.control(
                        num_ctrl_qubits=1,
                        method=self.ctrl_method,
                        ctrl_state=ctrl_state,
                    )

                    # Add condition truth value qubit to the instruction qubit list
                    instruction.qubits = [self.condition_truth_value] + list(
                        instruction.qubits
                    )
                # Append instruction
                self.env_qs.append(instruction)

            perm_unlock(ctrl_qubits)

            if inversion_tracker == -1:
                x(self.condition_truth_value)

            if len(ctrl_qubits) > 1:

                if len(ctrl_qubits) > 5:
                    method = "auto"
                else:
                    method = "gray_pt_inv"

                mcx(
                    ctrl_qubits,
                    self.qbool,
                    method=method,
                    ctrl_state=cond_compile_ctrl_state,
                )
                self.qbool.delete()

        if self.parent_cond_env is not None:
            self.parent_cond_env.sub_condition_envs.extend(
                self.sub_condition_envs + [self]
            )

    def jcompile(self, eqn, context_dic):

        from qrisp.jasp import extract_invalues, insert_outvalues

        args = extract_invalues(eqn, context_dic)
        body_jaspr = eqn.params["jaspr"]

        num_ctrl = len(args) - len(body_jaspr.invars)
        flattened_jaspr = body_jaspr.flatten_environments()
        controlled_jaspr = flattened_jaspr.control(num_ctrl, ctrl_state=self.ctrl_state)

        import jax

        res = jax.jit(controlled_jaspr.eval)(*args)

        # Retrieve the equation
        jit_eqn = get_last_equation()
        
        jit_eqn.params["jaxpr"] = controlled_jaspr
        jit_eqn.params["name"] = "ctrl_env"

        if not isinstance(res, tuple):
            res = (res,)

        insert_outvalues(eqn, context_dic, res)


# This function turns instructions where the definition contains CustomControlOperations
# into CustomControlOperations. For this it checks that all Instructions that are acting
# on the control_qubit are also CustomControlOperations.
# For the conversion process, this function turns all Operations which are NO custom
# controls into their regular controls.
def convert_to_custom_control(instruction, control_qubit, invert_control=False):

    from qrisp.environments import CustomControlOperation

    if invert_control:
        qc = QuantumCircuit(len(instruction.qubits))
        cusc_x = CustomControlOperation(XGate(), targeting_control=True)
        qc.append(cusc_x, [qc.qubits[instruction.qubits.index(control_qubit)]])
        qc.append(instruction.op, qc.qubits)
        qc.append(cusc_x, [qc.qubits[instruction.qubits.index(control_qubit)]])

        res = instruction.copy()
        res.op = qc.to_gate()

        return convert_to_custom_control(res, control_qubit)

    # If the Operation is already a CustomControlOperation, do nothing
    if isinstance(instruction.op, CustomControlOperation):
        return instruction

    # If the Operation is primitive, an error happened during compilation,
    # since Operations which are not CustomControlledOperations should not target
    # the control qubit
    if instruction.op.definition is None:
        print(instruction)
        raise Exception

    # We now generate the new Instruction
    new_definition = instruction.op.definition.clearcopy()
    new_control_qubit = new_definition.qubits[instruction.qubits.index(control_qubit)]

    # Iterate through the data
    for def_instr in instruction.op.definition.data:

        if new_control_qubit in def_instr.qubits:
            # If the instruction is targeting the control qubit, we call the function
            # recursively to make sure that we are indeed appending a custom_control
            # operation
            new_definition.append(
                convert_to_custom_control(def_instr, new_control_qubit)
            )
        else:
            # Else, we generate the operations regular control
            new_op = def_instr.op.control(1)
            new_definition.append(
                new_op, [new_control_qubit] + def_instr.qubits, def_instr.clbits
            )

    # Create the result and modify the definition
    res = instruction.copy()
    res.op.definition = new_definition

    res.op = CustomControlOperation(
        res.op, targeting_control=control_qubit in instruction.qubits
    )

    return res


def control(*args, **kwargs):
    args = list(args)
    from qrisp import Qubit, QuantumBool, QuantumVariable
    from qrisp.jasp import AbstractQubit, check_for_tracing_mode

    if isinstance(args[0], QuantumVariable):
        if isinstance(args[0], QuantumBool):
            args[0] = [args[0][0]]
        else:
            args[0] = list(args[0])
    if not isinstance(args[0], list):
        args[0] = [args[0]]

    if check_for_tracing_mode():
        if all(isinstance(obj, bool) for obj in [x for x in args[0]]):
            return ClControlEnvironment(*args, **kwargs)
        elif all(isinstance(obj, AbstractQubit) for obj in [x.aval for x in args[0]]):
            return ControlEnvironment(*args, **kwargs)
        elif all(isinstance(obj, ShapedArray) for obj in [x.aval for x in args[0]]):
            return ClControlEnvironment(*args, **kwargs)
        else:
            raise Exception(f"Don't know how to control from input type {args[0]}")
    else:
        if all(isinstance(obj, (Qubit, QuantumBool)) for obj in args[0]):
            return ControlEnvironment(*args, **kwargs)
        elif all(isinstance(obj, (bool, np.bool)) for obj in [x for x in args[0]]):
            return ClControlEnvironment(*args, **kwargs)
        elif all(isinstance(obj, ArrayImpl) for obj in [x for x in args[0]]):
            args[0] = [bool(bit) for bit in args[0]]
            return ClControlEnvironment(*args, **kwargs)
        else:
            raise Exception(f"Don't know how to control from input type {args[0]}")



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp.jasp.primitives import (
    create_qubits_p, 
    get_qubit_p, 
    get_size_p,
    slice_p,
    fuse_p,
    reset_p,
    Measurement_p,
    delete_qubits_p,
    create_quantum_kernel_p,
    consume_quantum_kernel_p,
    quantum_gate_p,
    AbstractQuantumCircuit,
    AbstractQubitArray,
    AbstractQubit)

import qrisp.jasp.mlir.dialect_implementation._jasp_ops_gen as jasp_dialect

from jax.interpreters.mlir import ir_type_handlers
from jaxlib.mlir import ir

##########################
# Register type lowering #
##########################

# These types need to be callable since the can only be 
# created when inside an MLIR Context manager
def get_ir_qst_type():
    return ir.Type.parse("!jasp.QuantumState")

def get_ir_qa_type():
    return ir.Type.parse("!jasp.QubitArray") 

def get_ir_qb_type():
    return ir.Type.parse("!jasp.Qubit")

# Define builtin integer type functions
def get_i64_type():
    return ir.IntegerType.get_signless(64)

def get_i1_type():
    return ir.IntegerType.get_signless(1)

# Register in ir_type_handlers

# Register type lowering (like Catalyst does)
def _aqc_lowering(aval):
    assert isinstance(aval, AbstractQuantumCircuit)
    return get_ir_qst_type()
def _aqa_lowering(aval):
    assert isinstance(aval, AbstractQubitArray)
    return get_ir_qa_type()
def _aqb_lowering(aval):
    assert isinstance(aval, AbstractQubit)
    return get_ir_qb_type()

# Register in ir_type_handlers
ir_type_handlers[AbstractQuantumCircuit] = _aqc_lowering
ir_type_handlers[AbstractQubitArray] = _aqa_lowering
ir_type_handlers[AbstractQubit] = _aqb_lowering


###############################
# Register Primitive lowering #
###############################

lowering_rules = []

def create_qubits_lowering(ctx, amount, qst_in):
    """
    Lowering rule that emits our CreateQubits dialect operation.
    """
    # Create our create_qubits operation using the generated class
    create_qubits_op = jasp_dialect.CreateQubitsOp(get_ir_qa_type(), get_ir_qst_type(), amount, qst_in)
    # Return both results: QubitArray and QuantumState as list
    return [create_qubits_op.results[0], create_qubits_op.results[1]]

lowering_rules.append((create_qubits_p, create_qubits_lowering))

def get_qubit_lowering(ctx, qb_array, position):
    """
    Lowering rule that emits our GetQubit dialect operation.
    """
    # Create our get_qubit operation using the generated class
    get_qubit_op = jasp_dialect.GetQubitOp(get_ir_qb_type(), qb_array, position)
    # Return the result of our operation
    return [get_qubit_op.results[0]]

lowering_rules.append((get_qubit_p, get_qubit_lowering))

def get_size_lowering(ctx, qb_array):
    """
    Lowering rule that emits our GetSize dialect operation.
    """
    # Create scalar tensor type for i64 result
    i64_result_type = ir.RankedTensorType.get([], get_i64_type())  # scalar tensor of i64
    
    # Create our get_size operation using the generated class
    get_size_op = jasp_dialect.GetSizeOp(i64_result_type, qb_array)
    # Return the result of our operation
    return [get_size_op.results[0]]

lowering_rules.append((get_size_p, get_size_lowering))

def slice_lowering(ctx, qb_array, start, end):
    """
    Lowering rule that emits our Slice dialect operation.
    """
    # Create our slice operation using the generated class
    slice_op = jasp_dialect.SliceOp(get_ir_qa_type(), qb_array, start, end)
    # Return the result QubitArray
    return [slice_op.results[0]]

lowering_rules.append((slice_p, slice_lowering))

def fuse_lowering(ctx, operand1, operand2):
    """
    Lowering rule that emits our Fuse dialect operation.
    """
    # Create our fuse operation using the generated class
    fuse_op = jasp_dialect.FuseOp(get_ir_qa_type(), operand1, operand2)
    # Return the result QubitArray
    return [fuse_op.results[0]]

lowering_rules.append((fuse_p, fuse_lowering))

def reset_lowering(ctx, qubits, in_qst):
    """
    Lowering rule that emits our Reset dialect operation.
    """
    # Create our reset operation using the generated class
    reset_op = jasp_dialect.ResetOp(get_ir_qst_type(), qubits, in_qst)
    # Return the result quantum state
    return [reset_op.results[0]]

lowering_rules.append((reset_p, reset_lowering))

def measure_lowering(ctx, meas_q, in_qst):
    """
    Lowering rule that emits our Measure dialect operation.
    """
    if meas_q.type == get_ir_qa_type():
        # For QubitArray measurement, result is tensor<i64>
        res_type = ir.RankedTensorType.get([], get_i64_type())  # scalar tensor of i64
    elif meas_q.type == get_ir_qb_type():
        # For single Qubit measurement, result is tensor<i1>
        res_type = ir.RankedTensorType.get([], get_i1_type())   # scalar tensor of i1
    else:
        raise Exception(f"Unknown qubit type: {meas_q.type}")
    
    # Create our measure operation using the generated class
    measure_op = jasp_dialect.MeasureOp(res_type, get_ir_qst_type(), meas_q, in_qst)
    # Return both results: measurement result and quantum state
    return [measure_op.results[0], measure_op.results[1]]

lowering_rules.append((Measurement_p, measure_lowering))

def delete_qubits_lowering(ctx, qubits, in_qst):
    """
    Lowering rule that emits our DeleteQubits dialect operation.
    """
    # Create our delete_qubits operation using the generated class
    delete_qubits_op = jasp_dialect.DeleteQubitsOp(get_ir_qst_type(), qubits, in_qst)
    # Return the result quantum state
    return [delete_qubits_op.results[0]]

lowering_rules.append((delete_qubits_p, delete_qubits_lowering))

def create_quantum_kernel_lowering(ctx):
    """
    Lowering rule that emits our CreateQuantumKernel dialect operation.
    """
    # Create our create_quantum_kernel operation using the generated class
    create_quantum_kernel_op = jasp_dialect.CreateQuantumKernelOp(get_ir_qst_type())
    # Return the result quantum state
    return [create_quantum_kernel_op.results[0]]

lowering_rules.append((create_quantum_kernel_p, create_quantum_kernel_lowering))

def consume_quantum_kernel_lowering(ctx, qst):
    """
    Lowering rule that emits our ConsumeQuantumKernel dialect operation.
    """
    # Create scalar tensor type for boolean result
    bool_result_type = ir.RankedTensorType.get([], get_i1_type())  # scalar tensor of i1
    
    # Create our consume_quantum_kernel operation using the generated class
    consume_quantum_kernel_op = jasp_dialect.ConsumeQuantumKernelOp(bool_result_type, qst)
    # Return the boolean result indicating success
    return [consume_quantum_kernel_op.results[0]]

lowering_rules.append((consume_quantum_kernel_p, consume_quantum_kernel_lowering))

def quantum_gate_lowering(ctx, *args, **params):
    """
    Lowering rule that emits our QuantumGate dialect operation.
    """
    
    # Extract gate type from params (JAX primitive parameters)
    op = params.get('gate', None)
    
    
    # Convert gate_type string to MLIR string attribute
    gate_type_attr = ir.StringAttr.get(op.name)
    
    
    # args contains: [gate_operands..., quantum_state]
    # Separate gate operands from the quantum state (last argument)
    gate_operands = args[:-1]  # All arguments except the last
    quantum_state = args[-1]   # Last argument is the quantum state
    
    # Create our quantum_gate operation using the generated class
    quantum_gate_op = jasp_dialect.QuantumGateOp(
        get_ir_qst_type(),           # Result type
        gate_type_attr,          # Gate type as string attribute
        gate_operands,           # Variadic gate operands (parameters + qubits)
        quantum_state            # Input quantum state
    )
    # Return the result quantum state
    return [quantum_gate_op.results[0]]

lowering_rules.append((quantum_gate_p, quantum_gate_lowering))

jasp_lowering_rules = tuple(lowering_rules)


"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np

from qrisp.alg_primitives import QFT
from qrisp.core.gate_application_functions import p, cz
from qrisp.environments import QuantumEnvironment, conjugate
from qrisp.circuit import Operation, PGate, QuantumCircuit


def fourier_adder(a, b, perform_QFT=True):
    """
    In-place adder function based on `this paper <https://arxiv.org/abs/quant-ph/0410184>`__.
    Performs the addition:

    ::

        b += a


    Parameters
    ----------
    a : int or QuantumVariable or list[Qubit]
        The value that should be added.
    b : QuantumVariable or list[Qubit]
        The value that should be modified in the in-place addition.
    perform_QFT : bool, optional
        If set to ``False``, no QFT is executed. The default is ``True``.

    Examples
    --------

    We add two integers:

    >>> from qrisp import QuantumFloat, fourier_adder
    >>> a = QuantumFloat(4)
    >>> b = QuantumFloat(4)
    >>> a[:] = 4
    >>> b[:] = 5
    >>> fourier_adder(a,b)
    >>> print(b)
    {9: 1.0}

    """

    if perform_QFT:
        env = conjugate(QFT)(b, exec_swap=False)
    else:
        env = QuantumEnvironment()

    with env:

        b = list(b)
        b = b[::-1]

        if isinstance(a, int):
            for i in range(len(b)):
                p(a * np.pi * 2 ** (1 + i - len(b)), b[i])

        else:

            if len(a) > len(b):
                raise Exception(
                    "Tried to add QuantumFloat of higher precision onto QuantumFloat of lower precision"
                )

            phase_correction_a = np.zeros(len(a))
            phase_correction_b = np.zeros(len(b))
            for j in range(len(a)):
                for i in range(len(b)):

                    if 1 + j + i - len(b) >= 1:
                        continue
                    if 1 + j + i - len(b) == 0:
                        cz(a[j], b[i])
                    else:
                        b[i].qs().append(
                            QuasiRZZ(-np.pi * 2 ** (1 + j + i - len(b)) / 2),
                            [a[j], b[i]],
                        )

                        phase_correction_a[j] += np.pi * 2 ** (1 + j + i - len(b)) / 2
                        phase_correction_b[i] += np.pi * 2 ** (1 + j + i - len(b)) / 2

            for i in range(len(b)):
                if phase_correction_b[i] % (2 * np.pi) != 0:
                    p(phase_correction_b[i], b[i])
            for i in range(len(a)):
                if phase_correction_a[i] % (2 * np.pi) != 0:
                    p(phase_correction_a[i], a[i])


class QuasiRZZ(Operation):

    def __init__(self, angle):
        qc = QuantumCircuit(2)
        qc.cx(qc.qubits[1], qc.qubits[0])
        qc.p(angle, qc.qubits[0])
        qc.cx(qc.qubits[1], qc.qubits[0])

        Operation.__init__(
            self, "quasi_rzz", num_qubits=2, definition=qc, params=[angle]
        )

        self.permeability = {0: True, 1: True}
        self.is_qfree = True

    def inverse(self):
        return QuasiRZZ(-self.params[0])

    def control(self, num_ctrl_qubits=1, ctrl_state=-1, method=None):

        qc = QuantumCircuit(2 + num_ctrl_qubits)
        qc.cx(qc.qubits[-1], qc.qubits[-2])
        qc.append(
            PGate(self.params[0]).control(
                num_ctrl_qubits, ctrl_state=ctrl_state, method=method
            ),
            qc.qubits[:-1],
        )
        qc.cx(qc.qubits[-1], qc.qubits[-2])

        res = Operation.control(
            self, num_ctrl_qubits, ctrl_state=ctrl_state, method=method
        )

        res.definition = qc
        return res



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np
import scipy as sp
from scipy.special import comb
import jax.numpy as jnp
from qrisp import (
    QuantumVariable,
    QuantumFloat,
    x,
    z,
    ry,
    control,
    conjugate,
    measure,
)
from qrisp.alg_primitives.reflection import reflection
from qrisp.alg_primitives.prepare import prepare
from qrisp.jasp import jrange, RUS
from qrisp.operators import QubitOperator


def CKS_parameters(A, eps, kappa=None, max_beta=None):
    """
    Computes the Chebyshev-series complexity parameter :math:`\\beta` and the truncation order :math:`j_0` for the
    truncated Chebyshev approximation of :math:`1/x` used in the `Childs–Kothari–Somma quantum algorithm <https://arxiv.org/abs/1511.02306>`_.

    To avoid confusion with the vector :math:`\ket{b}`, we represent the complexity parameter as :math:`\\beta`.

    This function distinguishes two input cases:

    - **Matrix input** 
        If ``A`` is either a NumPy array (Hermitian matrix), or a SciPy Compressed Sparse Row matrix (Hermitian matrix), the
        condition number :math:`\\kappa` is computed internally as :math:`\\kappa = \\mathrm{cond}(A)` if it is not specified.
    - **Block-encoding input** 
        If ``A`` is a 3-tuple representing a block-encoding, ``kappa`` must be specified.

    Given the condition number :math:`\\kappa` and the target precision
    :math:`\\epsilon`, the parameters are computed as:

    .. math::

        \\beta = \kappa^2 \log\!\left(\\frac{\kappa}{\epsilon}\\right),
        \quad
        j_0 = \sqrt{\\beta \log\!\left(\\frac{4\\beta}{\epsilon}\\right)}.

    If ``max_beta`` is provided, :math:`\\beta` is capped to
    :math:`\\min(\\beta, \\beta_{\\max})`. The returned values are cast to integers
    via the floor operation.

    Parameters
    ----------
    A : numpy.ndarray or scipy.sparse.csr_matrix or tuple
        Either the Hermitian matrix :math:`A` of size :math:`N \\times N` from
        the linear system :math:`A \\vec{x} = \\vec{b}`, or a 3-tuple
        representing a preconstructed block-encoding.
    eps : float
        Target precision :math:`\epsilon`, such that the prepared state :math:`\ket{\\tilde{x}}` is within
        :math:`\epsilon` of :math:`\ket{x}`.
    kappa : float, optional
        Condition number :math:`\\kappa` of :math:`A`. Required when ``A`` is
        a block-encoding tuple ``(U, state_prep, n)`` rather than a matrix.
    max_beta : float, optional
        Optional upper bound on the complexity parameter :math:`\\beta`.

    Returns
    -------
    j_0 : int
        Truncation order of the Chebyshev expansion
        :math:`j_0 = \\lfloor\sqrt{\\beta \log(4\\beta/\epsilon)}\\rfloor`.
    beta : float
        Complexity parameter :math:`\\beta = \\lfloor\kappa^2 \log(\kappa/\epsilon)\\rfloor`.
    """
    if kappa != None:
        # kappa is specified by the user
        pass 
    elif isinstance(A, tuple): # block-encoding
        raise Exception("Condition number must be specified if A is a block-encoding")
    else: # matrix
        from scipy.sparse import csr_matrix
        from numpy import ndarray

        if isinstance(A, ndarray):
            kappa = np.linalg.cond(A)
        elif isinstance(A, csr_matrix):
            lam_max = sp.sparse.linalg.eigsh(A, k=1, which='LA', return_eigenvectors=False)[0]
            lam_min = sp.sparse.linalg.eigsh(A, k=1, which='SA', return_eigenvectors=False)[0]
            kappa = float(abs(lam_max) / abs(lam_min))
        else:
            raise TypeError(f"Unsupported type {type(A)} for the matrix A")

    if max_beta == None:
        beta = kappa**2 * np.log(kappa / eps)
    else:
        beta = min(kappa**2 * np.log(kappa / eps), max_beta)
    j_0 = np.sqrt(beta * np.log(4 * beta / eps))
    return int(j_0), int(beta)


def cheb_coefficients(j0, b):
    """
    Calculates the positive coefficients :math:`\\alpha_i` for the truncated
    Chebyshev expansion of :math:`1/x` up to order :math:`2j_0+1`, as described in the `Childs–Kothari–Somma paper <https://arxiv.org/abs/1511.02306>`_.

    The approximation is expressed as a linear combination
    of odd Chebyshev polynomials truncated at index :math:`j_0` (Lemma 14):

    .. math::

        g(x) = 4 \\sum_{j=0}^{j_0} (-1)^j
        \\left[ \\sum_{i=j+1}^{b} \\frac{\\binom{2b}{b+i}}{2^{2b}} \\right]
        T_{2j+1}(x)

    The Linear Combination of Unitaries (LCU) lemma requires strictly
    positive coefficients :math:`\\alpha_i > 0`, their absolute values are
    used. The alternating factor :math:`(-1)^j` is later implemented as a set of
    Z-gates within the CKS circuit (see :func:`inner_CKS`).

    Parameters
    ----------
    j0 : int
        Truncation order of the Chebyshev expansion,
        :math:`j_0 = \\lfloor\sqrt{\\beta \log(4\\beta/\epsilon)}\\rfloor`.
    b : float
        Complexity parameter, :math:`\\beta = \\lfloor\kappa^2 \log(\kappa/\epsilon)\\rfloor`.

    Returns
    -------
    coeffs : numpy.ndarray
        Array of positive Chebyshev coefficients
        :math:`{\\alpha_{2j+1}}`, corresponding to the odd degrees of Chebyshev polynomials of the first kind :math:`T_1, T_3, \\dots, T_{2j_0+1}`.
    """
    coeffs = []
    for j in range(j0 + 1):
        sum_i = 0
        for i in range(j + 1, b + 1):
            sum_i += comb(2 * b, b + i)

        coeff = 4 * (2 ** (-2 * b)) * sum_i
        coeffs.append(coeff)
    return np.array(coeffs)


def unary_angles(coeffs):
    """
    Computes rotation angles :math:`\\phi_i` to prepare the unary state :math:`\ket{\\text{unary}}`,
    corresponding to the square-root-amplitude encoding of the Chebyshev coefficients.

    The prepared unary state is

    .. math::

        \ket{\\text{unary}} \propto
        \\sqrt{\\alpha_1}\ket{100\\dots00} + \\sqrt{\\alpha_3}\ket{110\\dots00} +
        \\cdots + \\sqrt{\\alpha_{2j_0+1}}\ket{111\\dots11}

    Parameters
    ----------
    coeffs : numpy.ndarray
        Positive Chebyshev coefficients :math:`\\alpha_i`.

    Returns
    -------
    phi : numpy.ndarray
        Rotation angles :math:`\\phi_i` for unary state preparation.
    """

    alpha = jnp.sqrt(coeffs) # coeffs need not to be normalized since unary angles only depend on their ratio
    phi = jnp.zeros(len(alpha) - 1)
    phi = phi.at[-1].set(
        jnp.arctan(alpha[-1] / alpha[-2])
    )  # Last angle is determined by ratio of final two amplitudes
    for i in range(len(phi) - 2, -1, -1):
        phi = phi.at[i].set(
            jnp.arctan(alpha[i + 1] / alpha[i] / jnp.cos(phi[i + 1]))
        )  # Recursively compute previous rotation angles
    return 2 * phi  # Compensate for factor 1/2 in ry


def unary_prep(case, coeffs):
    """
    Prepares the unary-encoded state :math:`\ket{\\text{unary}}` of Chebyshev coefficients.

    When applied to a variable in state :math:`\ket{0}`, the resulting superposition is

    .. math::

        \ket{\\text{unary}} \propto
        \\sqrt{\\alpha_1}\ket{100\\dots00} + \\sqrt{\\alpha_3}\ket{110\\dots00} +
        \\cdots + \\sqrt{\\alpha_{2j_0+1}}\ket{111\\dots11}

    Parameters
    ----------
    case : QuantumVariable
        Variable with :math:`j_0` qubits on which the unary state preparation will be performed.
        If the variable is in state :math:`\ket{0}`, the state :math:`\ket{\\text{unary}}` is prepared.
    coeffs : numpy.ndarray
        An array of :math:`j_0` Chebyshev coefficients :math:`\\alpha_1,\\alpha_3,\dotsc,\\alpha_{2j_0+1}`.
    
    """
    phi = unary_angles(coeffs)

    x(case[0])
    ry(phi[0], case[1])
    for i in jrange(1, case.size - 1):
        with control(case[i]):
            ry(phi[i], case[i + 1])


def inner_CKS(A, b, eps, kappa=None, max_beta=None):
    """
    Core implementation of the `Childs–Kothari–Somma (CKS) quantum algorithm <https://arxiv.org/abs/1511.02306>`_. 
    
    This function integrates core components of the CKS approach to construct the circuit:
    Chebyshev polynomial approximation, linear combination of unitaries (LCU), and
    `qubitization with reflection operators <https://arxiv.org/abs/2208.00567>`_. This function does
    not include the Repeat-Until-Success (RUS) protocol and thus serves as a subroutine
    for generating the underlying circuit of the CKS algorithm.

    The semantics of the approach can be illustrated with the following circuit schematics:

    .. image:: /_static/CKS_circuit.png
       :align: center

    Implementation overview:
      1. Compute the CKS parameters :math:`j_0` and :math:`\\beta` (:func:`CKS_parameters`).
      2. Generate Chebyshev coefficients and unary state angles
         (:func:`cheb_coefficients`, :func:`unary_prep`).
      3. Build the core LCU structure and qubitization operator (:func:`inner_CKS`).

    The goal of this algorithm is to apply the non-unitary operator :math:`A^{-1}` to the
    input state :math:`\ket{b}`. Following the Chebyshev approach introduced in the CKS paper, we express :math:`A^{-1}` as
    a linear combination of odd Chebyshev polynomials: 

    .. math::
    
        A^{-1}\propto\sum_{j=0}^{j_0}\\alpha_{2j+1}T_{2j+1}(A),

    where :math:`T_k(A)` are Chebyshev polynomials of the first kind and :math:`\\alpha_{2j+1} > 0` are computed
    via :func:`cheb_coefficients`. These operators can be efficiently implemented with qubitization, which relies on a unitary
    block encoding :math:`U` of the matrix :math:`A`, and a :ref:`reflection operator <reflection>` :math:`R`.
 
    .. note::

        Qubitization requires that the block encoding unitary :math:`U` is self-inverse (:math:`U^2=I`).

    The fundamental iteration step is defined as :math:`(RU)`, where :math:`R` reflects
    around the auxiliary block-encoding state :math:`\ket{G}`, prepared as the ``inner_case`` QuantumFloat.

    Repeated applications of these unitaries, :math:`(RU)^k`, yield a block encoding of the :math:`k`-th Chebyshev polynomial 
    of the first kind :math:`T_k(A)`.
    
    We then construct a linear combination of these block-encoded polynomials using the LCU structure 

    .. math::
    
        LCU\ket{0}\ket{\psi}=PREP^{\dagger}\cdot SEL\cdot PREP\ket{0}\ket{\psi}=\\tilde{A}\ket{0}\ket{\psi}.
    
    Here, the :math:`PREP` operation prepares an auxiliary ``out_case`` Quantumfloat in the unary state :math:`\ket{\\text{unary}}`
    that encodes the square root of the Chebyshev coefficients :math:`\sqrt{\\alpha_j}`. The :math:`SEL` operation selects and applies the
    appropriate Chebyshev polynomial operator :math:`T_k(A)`, implemented by :math:`(RU)^k`, controlled on ``out_case`` in the unary
    state :math:`\ket{\\text{unary}}`. Based on the Hamming-weight :math:`k` of :math:`\ket{\\text{unary}}`,
    the polynomial :math:`T_{2k-1}` is block encoded and applied to the circuit. 
    
    To construct a linear combination of Chebyshev polynomials up to the :math:`2j_0+1`-th order, as in the original paper, 
    our implementation requires :math:`j_0+1` qubits in the ``out_case`` state :math:`\ket{\\text{unary}}`.

    The Chebyshev coefficients alternate in sign :math:`(-1)^j\\alpha_j`.
    Since the LCU lemma requires :math:`\\alpha_j>0`, negative terms are accounted for
    by applying Z-gates on each index qubit in ``out_case``.

    This function supports interchangeable and independent input cases for both
    the matrix ``A`` and vector ``b`` from the QLSP :math:`A\\vec{x} = \\vec{b}`:

    **Case distinctions for A**

    - Matrix input: 
        ``A`` is either a NumPy array (Hermitian matrix), or SciPy Compressed Sparse Row matrix (Hermitian matrix).
        The block encoding unitary :math:`SEL` and state preparation unitary
        :math:`PREP` are constructed internally from ``A`` via
        :meth:`Pauli decomposition <qrisp.operators.qubit.QubitOperator.pauli_block_encoding>`.

    - Block-encoding input: 
        ``A`` is a 3-tuple ``(U, state_prep, n)`` representing a block-encoding: 
        
        * U : callable
            Callable ``U(case, operand)`` applies the block-encoding unitary :math:`SEL`.
        * state_prep : callable
            Callable ``state_prep(case)`` applies the block-encoding state preparatiion unitary :math:`PREP`
            to an auxiliary ``case`` QuantumVariable .
        * n : int
            The size of the auxiliary ``case`` QuantumVariable.
            
        In this case, (an upper bound for) the condition number ``kappa`` must be specified.
        Additionally, the block-encoding unitary :math:`U` supplied must satisfy
        the property :math:`U^2 = I`, i.e., it is self-inverse. This condition is
        required for the correctness of the Chebyshev polynomial block encoding
        and qubitization step. Further details can be found `here <https://arxiv.org/abs/2208.00567>`_.

    **Case distinctions for b**

    - Vector input: 
        If ``b`` is a NumPy array, a new ``operand`` QuantumFloat is constructed, and state preparation is performed via
        ``prepare(operand, b)``.

    - Callable input:
        If ``b`` is a callable, it is assumed to prepare the operand state when called. 
        The ``operand`` QuantumFloat is generated directly via ``operand = b()``.

    Parameters
    ----------
    A : numpy.ndarray or scipy.sparse.csr_matrix or tuple
        Either the Hermitian matrix :math:`A` of size :math:`N \\times N` from
        the linear system :math:`A \\vec{x} = \\vec{b}`, or a 3-tuple
        ``(U, state_prep, n)`` representing a preconstructed block-encoding.
    b : numpy.ndarray or callable
        Vector :math:`\\vec{b}` of the linear system, or a
        callable that prepares the corresponding quantum state.
    eps : float
        Target precision :math:`\epsilon`, such that the prepared state :math:`\ket{\\tilde{x}}` is within error
        :math:`\epsilon` of :math:`\ket{x}`.
    kappa : float, optional
        Condition number :math:`\\kappa` of :math:`A`. Required when ``A`` is
        a tuple (block-encoding) rather than a matrix.
    max_beta : float, optional
        Optional upper bound on the complexity parameter :math:`\\beta`.

    Returns
    -------
    operand : QuantumVariable
        Operand variable after applying the LCU protocol.
    in_case : QuantumFloat
        Auxiliary variable after applying the LCU protocol.
        Must be measured in state $\ket{0}$ for the LCU protocol to be successful.
    out_case : QuantumFloat
        Auxiliary variable after applying the LCU protocol.
        Must be measured in state $\ket{0}$ for the LCU protocol to be successful.
        
    Examples
    --------

    The example shows how to solve a QLSP using the ``inner_CKS`` function directly,
    without employing the :ref:`RUS` protocol in Qrisp.

    First, define a Hermitian matrix :math:`A` and a right-hand side vector :math:`\\vec{b}`

    ::

        import numpy as np

        A = np.array([[0.73255474, 0.14516978, -0.14510851, -0.0391581],
                      [0.14516978, 0.68701415, -0.04929867, -0.00999921],
                      [-0.14510851, -0.04929867, 0.76587818, -0.03420339],
                      [-0.0391581, -0.00999921, -0.03420339, 0.58862043]])

        b = np.array([0, 1, 1, 1])

    Then, constuct the CKS circuit and perform a multi measurement: 

    ::

        from qrisp.algorithms.cks import inner_CKS
        from qrisp import multi_measurement

        operand, in_case, out_case = inner_CKS(A, b, 0.001)
        res_dict = multi_measurement([operand, in_case, out_case])

    This performs the measurement on all three of our QuantumVariables ``out_case``, ``in_case``, and ``operand``.
    Since the CKS (and all other LCU based approaches) is correctly performed only when
    auxiliary QuantumVariables are measured in :math:`\ket{0}`, we need
    to construct a new dictionary, and collect the measurements when this is true.

    ::

        new_dict = dict()
        success_prob = 0

        for key, prob in res_dict.items():
            if key[1]==0 and key[2]==0:
                new_dict[key[0]] = prob
                success_prob += prob

        for key in new_dict.keys():
            new_dict[key] = new_dict[key]/success_prob

        for k, v in new_dict.items():
            new_dict[k] = v**0.5
    
    Finally, compare the quantum simulation result with the classical solution:

    ::

        q = np.array([new_dict.get(key, 0) for key in range(len(b))])
        c = (np.linalg.inv(A) @ b) / np.linalg.norm(np.linalg.inv(A) @ b)

        print("QUANTUM SIMULATION\\n", q, "\\nCLASSICAL SOLUTION\\n", c)
        # QUANTUM SIMULATION
        # [0.0289281  0.55469649 0.53006773 0.64070521] 
        # CLASSICAL SOLUTION
        # [0.02944539 0.55423278 0.53013239 0.64102936]

    This approach enables execution of the compiled CKS circuit on :ref:`compatible quantum
    hardware or simulators <BackendInterface>`.  

    """
    if isinstance(A, tuple) and len(A) == 3:
        U, state_prep, n = A
    else:
        H = QubitOperator.from_matrix(A, reverse_endianness=True)
        U, state_prep, n = (H.pauli_block_encoding())  # Construct block encoding of A as a set of Pauli unitaries

    j_0, beta = CKS_parameters(A, eps, kappa, max_beta)
    cheb_coeffs = cheb_coefficients(j_0, beta)

    def RU(case, operand):
        """
        Applies one qubitization step :math:`RU` (or its inverse) associated
        with the Chebyshev polynomial block-encoding of an operator :math:`A`.

        This operation alternates between application of the block-encoding
        unitary :math:`U` and a reflection about the block-encoding state
        :math:`\ket{G}`, realizing

        .. math::

            T_k(A) = (RU)^k,

        where :math:`R` reflects about the block-encoding state :math:`\ket{G}`,
        and :math:`T_k(A)` is the Chebyshev polynomials of the first kind of degree :math:`k`.

        Parameters
        ----------
        case : QuantumVariable
            Auxiliary case variable encoding the block-encoding state :math:`\ket{G}`.
        operand : QuantumVariable
            Operand (solution) variable upon which the block-encoded matrix
            :math:`A` acts. 

        """
        U(case, operand)
        reflection(case, state_function=state_prep)  # reflection operator R about $\ket{G}$.

    out_case = QuantumFloat(j_0 + 1)
    in_case = QuantumFloat(n)

    if callable(b):
        operand = b()
    else:
        operand = QuantumFloat(int(np.log2(b.shape[0])))
        prepare(operand, b)

    # Core LCU protocol: PREP, SELECT, PREP^†
    with conjugate(unary_prep)(out_case, cheb_coeffs):
        with conjugate(state_prep)(in_case):
            with control(out_case[0]):
                RU(in_case, operand)
            for i in jrange(1, j_0 + 1):
                z(out_case[i])
                with control(out_case[i]):
                    RU(in_case, operand)
                    RU(in_case, operand)

    return operand, in_case, out_case

def inner_CKS_wrapper(qlsp, eps, kappa=None, max_beta=None):
    """
    Wrapper for the full Childs–Kothari–Somma (CKS) algorithm, performing post-selection using the RUS (Repeat-Until-Success) protocol.

    This function calls a supplied problem generator ``qlsp`` to obtain the linear system,
    constructs the required circuit via :func:`inner_CKS`, and returns the solution operand QuantumFloat
    if post-selection on auxiliary QuantumFloats succeeds.

    The algorithm succeeds if both auxiliary QuantumFloats ``in_case`` and ``out_case`` are measured
    in the :math:`\ket{0}` state. If post-selection fails, the simulation
    restarts until successful as managed by the :func:`RUS` decorator.

    Parameters
    ----------
    qlsp : callable
        Function returning a tuple :math:`(A, \\vec{b})`, where ``A`` is either a Hermitian matrix
        or a 3-tuple with preconstructed block-encoding (see :func:`inner_CKS`), and ``b`` is either
        a NumPy array or a callable that prepares a quantum state.
    eps : float
        Target precision :math:`\epsilon`, such that the prepared state :math:`\ket{\\tilde{x}}` is within
        :math:`\epsilon` of :math:`\ket{x}.
    kappa : float, optional
        Condition number :math:`\\kappa` of :math:`A`. Required when ``A`` is
        a block-encoding tuple ``(U, state_prep, n)`` rather than a matrix.
    max_beta : float, optional
        Optional upper bound on the complexity parameter :math:`\\beta`.

    Returns
    -------
    success_bool : bool
        Indicates whether post-selection succeeded. Only when ``success_bool`` is ``True``
        does ``operand`` contain the proper solution state; otherwise, the simulation
        will run again according to the RUS protocol.
    operand : QuantumVariable
        QuantumFloat containing the resulting (approximate) solution state
        :math:`\ket{\\tilde x} \\propto A^{-1}\ket{b}` if ``success_bool`` is ``True``.
    """

    A, b = qlsp()

    operand, in_case, out_case = inner_CKS(A, b, eps, kappa, max_beta)

    success_bool = (measure(out_case) == 0) & (measure(in_case) == 0)

    return success_bool, operand


# Apply the RUS decorator with the workaround in order to show in documentation
temp_docstring = inner_CKS_wrapper.__doc__
inner_CKS_wrapper = RUS(static_argnums=[1, 2])(inner_CKS_wrapper)
inner_CKS_wrapper.__doc__ = temp_docstring


def CKS(A, b, eps, kappa=None, max_beta=None):
    """
    Performs the `Childs–Kothari–Somma (CKS) quantum algorithm <https://arxiv.org/abs/1511.02306>`_ to solve the Quantum Linear Systems Problem (QLSP)
    :math:`A \\vec{x} = \\vec{b}`, using the Chebyshev approximation of :math:`1/x`. 
    
    The algorithm prepares a quantum state :math:`\ket{x} \propto A^{-1} \ket{b}`
    within target precision :math:`\epsilon` of the ideal solution :math:`\ket{x}`. 
    
    The asymptotic complexity is
    :math:`\\mathcal{O}\\!\\left(\\log(N)s \\kappa^2 \\text{polylog}\\!\\frac{s\\kappa}{\\epsilon}\\right)`, where :math:`N` is the matrix size, :math:`s` its sparsity, and 
    :math:`\kappa` its condition number. This represents an exponentially
    better precision scaling compared to the HHL algorithm.

    This function integrates all core components of the CKS approach:
    Chebyshev polynomial approximation, linear combination of unitaries (LCU),
    `qubitization with reflection operators <https://arxiv.org/abs/2208.00567>`_, and the :ref:`Repeat-Until-Success (RUS) <RUS>` protocol.
    The semantics of the approach can be illustrated with the following circuit schematics:

    .. image:: /_static/CKS_schematics.png
       :align: center

    Implementation overview:
      1. Compute the CKS parameters :math:`j_0` and :math:`\\beta` (:func:`CKS_parameters`).
      2. Generate Chebyshev coefficients and the auxiliary unary state (:func:`cheb_coefficients`, :func:`unary_prep`).
      3. Build the core LCU structure and qubitization operator (:func:`inner_CKS`).
      4. Apply the RUS post-selection step to project onto the valid success outcome.

    This function supports interchangeable and independent input types for both
    the matrix ``A`` and vector ``b`` from the QLSP :math:`A\\vec{x} = \\vec{b}`.

    **Case distinctions for A**

    - Matrix input: 
        ``A`` is either a NumPy array (Hermitian matrix), or SciPy Compressed Sparse Row matrix (Hermitian matrix).
        The block encoding unitary :math:`SEL` and state preparation unitary
        :math:`PREP` are constructed internally from ``A`` via
        :meth:`Pauli decomposition <qrisp.operators.qubit.QubitOperator.pauli_block_encoding>`.

    - Block-encoding input: 
        ``A`` is a 3-tuple ``(U, state_prep, n)`` representing a block-encoding: 
        
        * U : callable
            Callable ``U(operand, case)`` applies the block-encoding unitary :math:`SEL`.
        * state_prep : callable
            Callable ``state_prep(case)`` applies the block-encoding state preparatiion unitary :math:`PREP`
            to an auxiliary ``case`` QuantumVariable .
        * n : int
            The size of the auxiliary ``case`` QuantumVariable.
            
        In this case, (an upper bound for) the condition number ``kappa`` must be specified.
        Additionally, the block-encoding unitary :math:`U` supplied must satisfy
        the property :math:`U^2 = I`, i.e., it is self-inverse. This condition is
        required for the correctness of the Chebyshev polynomial block encoding
        and qubitization step. Further details can be found `here <https://arxiv.org/abs/2208.00567>`_.

    **Case distinctions for b**

    - Vector input: 
        If ``b`` is a NumPy array, a new ``operand`` QuantumFloat is constructed, and state preparation is performed via
        ``prepare(operand, b)``.

    - Callable input:
        If ``b`` is a callable, it is assumed to prepare the operand state when called. 
        The ``operand`` QuantumFloat is generated directly via ``operand = b()``.

    Parameters
    ----------
    A : numpy.ndarray or scipy.sparse.csr_matrix or tuple
        Either the Hermitian matrix :math:`A` of size :math:`N \\times N` from
        the linear system :math:`A \\vec{x} = \\vec{b}`, or a 3-tuple
        ``(U, state_prep, n)`` representing a preconstructed block-encoding.
    b : numpy.ndarray or callable
        Either a vector :math:`\\vec{b}` of the linear system, or a
        callable that prepares the corresponding quantum state ``operand``.
    eps : float
        Target precision :math:`\epsilon`, such that the prepared state :math:`\ket{\\tilde{x}}` is within error
        :math:`\epsilon` of :math:`\ket{x}`.
    kappa : float, optional
        Condition number :math:`\\kappa` of :math:`A`. Required when ``A`` is
        a block-encoding tuple ``(U, state_prep, n)`` rather than a matrix.
    max_beta : float, optional
        Optional upper bound on the complexity parameter :math:`\\beta`.

    Returns
    -------
    operand : QuantumVariable
        Quantum variable containing the final (approximate) solution state
        :math:`\ket{\\tilde{x}} \propto A^{-1}\ket{b}`. When the internal
        :ref:`RUS <RUS>` decorator reports success (``success_bool = True``), this
        variable contains the valid post-selected solution. If ``success_bool``
        is ``False``, the simulation automatically repeats until success.
        
    Examples
    --------

    The following examples demonstrate how the CKS algorithm can be applied to solve the
    quantum linear systems problem (QLSP)
    :math:`A \\vec{x} = \\vec{b}`, using either a direct Hermitian matrix input or
    a preconstructed block-encoding representation.
    
    **Example 1: Solving a 4×4 Hermitian system**

    First, we define a small Hermitian matrix :math:`A` and a right-hand side vector :math:`\\vec{b}`:

    ::

        import numpy as np

        A = np.array([[0.73255474, 0.14516978, -0.14510851, -0.0391581],
                      [0.14516978, 0.68701415, -0.04929867, -0.00999921],
                      [-0.14510851, -0.04929867, 0.76587818, -0.03420339],
                      [-0.0391581, -0.00999921, -0.03420339, 0.58862043]])

        b = np.array([0, 1, 1, 1])

    Next, we solve this linear system using the CKS quantum algorithm:

    ::

        from qrisp.algorithms.cks import CKS
        from qrisp.jasp import terminal_sampling

        @terminal_sampling
        def main():

            x = CKS(A, b, 0.01)
            return x

        res_dict = main()

    The resulting dictionary contains the measurement probabilities for each computational basis state.
    To extract the corresponding quantum amplitudes (up to sign):

    ::

        for k, v in res_dict.items():
            res_dict[k] = v**0.5

        q = np.array([res_dict.get(key, 0) for key in range(len(b))])
        print("QUANTUM SIMULATION\\n", q)
        # QUANTUM SIMULATION
        # [0.02714082 0.55709921 0.53035395 0.63845794]
    
    We compare this to the classical normalized solution:

    ::

        c = (np.linalg.inv(A) @ b) / np.linalg.norm(np.linalg.inv(A) @ b)

        print("CLASSICAL SOLUTION\\n", c)  
        # CLASSICAL SOLUTION
        # [0.02944539 0.55423278 0.53013239 0.64102936]

    We see that we obtained the same result in our quantum simulation up to precision :math:`\epsilon`!

    To perform quantum resource estimation, replace the ``@terminal_sampling``
    decorator with ``@count_ops(meas_behavior="0")``:

    ::
        
        from qrisp.jasp import count_ops
    
        @count_ops(meas_behavior="0")
        def main():

            x = CKS(A, b, 0.01)
            return x

        res_dict = main()
        print(res_dict)
        # {'cx': 3628, 't': 1288, 't_dg': 1702, 'p': 178, 'cz': 138, 'cy': 46, 'h': 960, 'x': 324, 's': 66, 's_dg': 66, 'u3': 723, 'gphase': 49, 'ry': 2, 'z': 11, 'measure': 16}

    The printed dictionary lists the estimated quantum gate counts required for the CKS algorithm. Since the simulation itself is not executed, 
    this approach enables scalable gate count estimation for linear systems of arbitrary size.
    
    **Example 2: Using a custom block encoding**

    The previous example displays how to solve the linear system for any Hermitian matrix :math:`A`,
    where the block-encoding is constructed from the Pauli decomposition of :math:`A`. While this approach is efficient when :math:`A`
    corresponds to, e.g., an Ising or a Heisenberg Hamiltonian, the number of Pauli terms may not scale favorably in general.
    For certain sparse matrices, their structure can be exploited to construct more efficient block-encodings.

    In this example, we construct a block-encoding representation for the following tridiagonal sparse matrix:

    ::

        import numpy as np

        def tridiagonal_shifted(n, mu=1.0, dtype=float):
            I = np.eye(n, dtype=dtype)
            return (2 + mu) * I - 2*np.eye(n, k=n//2, dtype=dtype) - 2*np.eye(n, k=-n//2, dtype=dtype)

        N = 8
        A = tridiagonal_shifted(N, mu=3)
        b = np.random.randint(0, 2, size=N)

        print(A)
        # [[ 5.  0.  0.  0. -2.  0.  0.  0.]
        #  [ 0.  5.  0.  0.  0. -2.  0.  0.]
        #  [ 0.  0.  5.  0.  0.  0. -2.  0.]
        #  [ 0.  0.  0.  5.  0.  0.  0. -2.]
        #  [-2.  0.  0.  0.  5.  0.  0.  0.]
        #  [ 0. -2.  0.  0.  0.  5.  0.  0.]
        #  [ 0.  0. -2.  0.  0.  0.  5.  0.]
        #  [ 0.  0.  0. -2.  0.  0.  0.  5.]]

    This matrix can be decomposed using three unitaries: the identity $I$, and two shift operators $V\colon\ket{k}\\rightarrow-\ket{k+N/2 \mod N}$ and $V^{\dagger}\colon\ket{k}\\rightarrow-\ket{k-N/2 \mod N}$.
    We define their corresponding functions:

    ::

        from qrisp import gphase, prepare, qswitch

        def I(qv):
            pass

        def V(qv):
            qv += N//2
            gphase(np.pi, qv[0])

        def V_dg(qv):
            qv -= N//2
            gphase(np.pi, qv[0])

        unitaries = [I, V, V_dg]

    Additionally, the block encoding unitary :math:`U` supplied must satisfy the property :math:`U^2 = I`, i.e., it is self-inverse.
    This condition is required for the correctness of the Chebyshev polynomial block encoding
    and qubitization step. Further details can be found `here <https://arxiv.org/abs/2208.00567>`_. In this case, the fact that $V^2=(V^{\dagger})^2=I$
    ensures that defining the block encoding unitary via a :ref:`quantum switch case <qswitch>` satsifies :math:`U^2 = I`.

    We now define the block_encoding ``(U, state_prep, n)``:

    ::

        coeffs = np.array([5,1,1])
        alpha = np.sum(coeffs)

        def U(case, operand):
            qswitch(operand, case, unitaries)

        def state_prep(case):
            prepare(case, np.sqrt(coeffs/alpha))

        block_encoding = (U, state_prep, 2)

    We solve the linear system by passing this block-encoding tuple as ``A`` into the CKS function:

    ::

        @terminal_sampling
        def main():

            x = CKS(block_encoding, b, 0.01, np.linalg.cond(A))
            return x

        res_dict = main()

    We, again, compare the solution obtained by quantum simulation with the classical solution

    ::

        for k, v in res_dict.items():
            res_dict[k] = v**0.5

        q = np.array([res_dict.get(key, 0) for key in range(N)])
        c = (np.linalg.inv(A) @ b) / np.linalg.norm(np.linalg.inv(A) @ b)

        print("QUANTUM SIMULATION\\n", q, "\\nCLASSICAL SOLUTION\\n", c)
        # QUANTUM SIMULATION
        # [0.43917486 0.31395935 0.12522139 0.43917505 0.43917459 0.12522104 0.31395943 0.43917502]
        # CLASSICAL SOLUTION
        # [0.43921906 0.3137279  0.12549116 0.43921906 0.43921906 0.12549116 0.3137279  0.43921906]

    Quantum resource estimation can be performed in the same way as in the first example:

    ::

        @count_ops(meas_behavior="0")
        def main():

            x = CKS(block_encoding, b, 0.01, np.linalg.cond(A))
            return x
    
        res_dict = main()
        print(res_dict)
        # {'h': 676, 't_dg': 806, 'cx': 1984, 't': 651, 's': 152, 's_dg': 90, 'u3': 199, 'gphase': 65, 'p': 149, 'z': 15, 'x': 126, 'measure': 142, 'ry': 2}

    Finally, note that it is also possible to solve a linear system without using the :ref:`RUS` decorator,
    by performing the necessary post-selection manually. Such an example can be found in :func:`inner_CKS`.

    """

    def qlsp():
        return A, b

    operand = inner_CKS_wrapper(qlsp, eps, kappa, max_beta)
    return operand



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import jax.numpy as jnp
from jaxlib.xla_extension import ArrayImpl
from jax import jit

from qrisp.jasp.tracing_logic import check_for_tracing_mode


class JRangeIterator:

    def __init__(self, *args):

        # Differentiate between the 3 possible cases of input signature

        if len(args) == 1:
            # In the case of one input argument, this argument is the stop value
            self.start = None
            self.stop = jnp.asarray(args[0], dtype="int64")
            self.step = jnp.asarray(1, dtype="int64")
        elif len(args) == 2:
            self.start = jnp.asarray(args[0], dtype="int64")
            self.stop = jnp.asarray(args[1], dtype="int64")
            self.step = jnp.asarray(1, dtype="int64")
        elif len(args) == 3:
            # Three arguments denote the case of a non-trivial step
            self.start = jnp.asarray(args[0], dtype="int64")
            self.stop = jnp.asarray(args[1], dtype="int64")
            self.step = jnp.asarray(args[2], dtype="int64")

        # The loop index should be inclusive because this makes loop inversion
        # much easier. For more details check inv_transform.py.
        self.stop -= 1

    def __iter__(self):
        self.iteration = 0

        # We create the loop iteration index tracer
        if self.start is None:
            self.loop_index = self.stop - self.stop
        else:
            self.loop_index = self.start + 0
        return self

    def __next__(self):
        # The idea is now to trace two iterations to capture what values get
        # updated after each iteration.
        # We capture the loop semantics using the JIterationEnvironment.
        # The actual jax loop primitive is then compiled in
        # JIterationEnvironment.jcompile
        from qrisp.jasp import TracingQuantumSession
        from qrisp import reset

        self.iteration += 1
        if self.iteration == 1:
            from qrisp.environments import JIterationEnvironment

            self.iter_env = JIterationEnvironment()
            # Enter the environment
            self.iter_env.__enter__()

            # We perform a trivial addition on the loop cancelation index.
            # This way the loop cancelation index will appear in the collected
            # quantum environment jaxpr and can therefore be identified as such.
            self.stop + 0

            self.iter_1_qvs = list(TracingQuantumSession.get_instance().qv_list)

            return self.loop_index

        elif self.iteration == 2:

            qs = TracingQuantumSession.get_instance()
            created_qvs = set(list(qs.qv_list)) - set(self.iter_1_qvs)
            created_qvs = list(created_qvs)
            created_qvs = sorted(created_qvs, key=lambda x: hash(x))

            if qs.gc_mode == "auto":
                for qv in created_qvs:
                    reset(qv)
                    qv.delete()
            elif qs.gc_mode == "debug" and len(created_qvs):
                raise Exception(
                    f"QuantumVariables {created_qvs} went out of scope without deletion during jrange"
                )

            # Perform the incrementation
            self.loop_index += self.step

            # Exit the old environment and enter the new one.
            self.iter_env.__exit__(None, None, None)
            self.iter_env.__enter__()
            # Similar to the incrementation above

            self.stop + 0
            self.iter_2_qvs = list(TracingQuantumSession.get_instance().qv_list)

            return self.loop_index

        elif self.iteration == 3:

            qs = TracingQuantumSession.get_instance()
            created_qvs = set(list(qs.get_instance().qv_list)) - set(self.iter_2_qvs)
            created_qvs = list(created_qvs)
            created_qvs = sorted(created_qvs, key=lambda x: hash(x))

            if qs.gc_mode == "auto":
                for qv in created_qvs:
                    reset(qv)
                    qv.delete()
            elif qs.gc_mode == "debug" and len(created_qvs):
                raise Exception(
                    f"QuantumVariables {created_qvs} went out of scope without deletion during jrange"
                )

            self.loop_index += self.step

            self.iter_env.__exit__(None, None, None)
            raise StopIteration


def jrange(*args):
    """
    Performs a loop with a dynamic bound. Similar to the Python native ``range``,
    this iterator can receive multiple arguments. If it receives just one, this
    value is interpreted as the stop value and the start value is assumed to be 0.
    Two arguments represent start and stop value, whereas three represent start,
    stop and step.

    .. warning::

        Similar to the :ref:`ClControlEnvironment <ClControlEnvironment>`, this feature must not have
        external carry values, implying values computed within the loop can't
        be used outside of the loop. It is however possible to carry on values
        from the previous iteration.

    .. warning::

        Each loop iteration must perform exactly the same instructions - the only
        thing that changes is the loop index


    Parameters
    ----------
    start : int
        The loop index to start at.
    stop : int
        The loop index to stop at.
    step : int
        The value to increase the loop index by after each iteration.

    Examples
    --------

    We construct a function that encodes an integer into an arbitrarily sized
    :ref:`QuantumVariable`:

    ::

        from qrisp import QuantumFloat, control, x
        from qrisp.jasp import jrange, make_jaspr

        @qache
        def int_encoder(qv, encoding_int):

            for i in jrange(qv.size):
                with control(encoding_int & (1<<i)):
                    x(qv[i])

        def test_f(a, b):

            qv = QuantumFloat(a)

            int_encoder(qv, b+1)

            return measure(qv)

        jaspr = make_jaspr(test_f)(1,1)

    Test the result:

    >>> jaspr(5, 8)
    9
    >>> jaspr(5, 9)
    10

    We now give examples that violate the above rules (ie. no carries and changing
    iteration behavior).

    To create a loop with carry behavior we simply return the final loop index

    ::

        @qache
        def int_encoder(qv, encoding_int):

            for i in jrange(qv.size):
                with control(encoding_int & (1<<i)):
                    x(qv[i])
            return i


        def test_f(a, b):

            qv = QuantumFloat(a)

            int_encoder(qv, b+1)

            return measure(qv)

        jaspr = make_jaspr(test_f)(1,1)

    >>> jaspr(5, 8)
    Exception: Found jrange with external carry value

    To demonstrate the second kind of illegal behavior, we construct a loop
    that behaves differently on the first iteration:

    ::

        @qache
        def int_encoder(qv, encoding_int):

            flag = True
            for i in jrange(qv.size):
                if flag:
                    with control(encoding_int & (1<<i)):
                        x(qv[i])
                else:
                    x(qv[0])
                flag = False

        def test_f(a, b):

            qv = QuantumFloat(a)

            int_encoder(qv, b+1)

            return measure(qv)

        jaspr = make_jaspr(test_f)(1,1)

    In this script, ``int_encoder`` defines a boolean flag that changes the
    semantics of the iteration behavior. After the first iteration the flag
    is set to ``False`` such that the alternate behavior is activated.

    >>> jaspr(5, 8)
    Exception: Jax semantics changed during jrange iteration

    """

    new_args = []
    if check_for_tracing_mode():
        for i in range(len(args)):
            if i == 2:
                new_args.append(args[i])
                continue
            if isinstance(args[i], (int, ArrayImpl)):
                new_args.append(make_tracer(args[i]))
            else:
                new_args.append(args[i])

        return JRangeIterator(*new_args)

    else:
        for i in range(len(args)):
            if not isinstance(args[i], int):
                new_args.append(int(args[i]))
            else:
                new_args.append(args[i])

        return range(*new_args)


def make_tracer(x):
    if isinstance(x, bool):
        dtype = jnp.bool
    elif isinstance(x, int):
        dtype = jnp.int64
    elif isinstance(x, float):
        dtype = jnp.float64
    elif isinstance(x, complex):
        dtype = jnp.complex32
    else:
        raise Exception(f"Don't know how to tracerize type {type(x)}")

    def tracerizer():
        return jnp.array(x, dtype)

    return jit(tracerizer)()


def jlen(x):
    if isinstance(x, list):
        return len(x)
    else:
        return x.size




# Autogenerated by mlir-tblgen; don't manually edit.

from ._ods_common import _cext as _ods_cext
from ._ods_common import (
    equally_sized_accessor as _ods_equally_sized_accessor,
    get_default_loc_context as _ods_get_default_loc_context,
    get_op_result_or_op_results as _get_op_result_or_op_results,
    get_op_results_or_values as _get_op_results_or_values,
    segmented_accessor as _ods_segmented_accessor,
)
_ods_ir = _ods_cext.ir

import builtins
from typing import Sequence as _Sequence, Union as _Union


@_ods_cext.register_dialect
class _Dialect(_ods_ir.Dialect):
  DIALECT_NAMESPACE = "jasp"

@_ods_cext.register_operation(_Dialect)
class ConsumeQuantumKernelOp(_ods_ir.OpView):
  OPERATION_NAME = "jasp.consume_quantum_kernel"

  _ODS_REGIONS = (0, True)

  def __init__(self, success, qst, *, loc=None, ip=None):
    operands = []
    results = []
    attributes = {}
    regions = None
    operands.append(qst)
    _ods_context = _ods_get_default_loc_context(loc)
    results.append(success)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def qst(self):
    return self.operation.operands[0]

  @builtins.property
  def success(self):
    return self.operation.results[0]

def consume_quantum_kernel(success, qst, *, loc=None, ip=None) -> _ods_ir.Value:
  return ConsumeQuantumKernelOp(success=success, qst=qst, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class CreateQuantumKernelOp(_ods_ir.OpView):
  OPERATION_NAME = "jasp.create_quantum_kernel"

  _ODS_REGIONS = (0, True)

  def __init__(self, result, *, loc=None, ip=None):
    operands = []
    results = []
    attributes = {}
    regions = None
    _ods_context = _ods_get_default_loc_context(loc)
    results.append(result)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def result(self):
    return self.operation.results[0]

def create_quantum_kernel(result, *, loc=None, ip=None) -> _ods_ir.Value:
  return CreateQuantumKernelOp(result=result, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class CreateQubitsOp(_ods_ir.OpView):
  OPERATION_NAME = "jasp.create_qubits"

  _ODS_REGIONS = (0, True)

  def __init__(self, result, qst_out, amount, qst_in, *, loc=None, ip=None):
    operands = []
    results = []
    attributes = {}
    regions = None
    operands.append(amount)
    operands.append(qst_in)
    _ods_context = _ods_get_default_loc_context(loc)
    results.append(result)
    results.append(qst_out)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def amount(self):
    return self.operation.operands[0]

  @builtins.property
  def qst_in(self):
    return self.operation.operands[1]

  @builtins.property
  def result(self):
    return self.operation.results[0]

  @builtins.property
  def qst_out(self):
    return self.operation.results[1]

def create_qubits(result, qst_out, amount, qst_in, *, loc=None, ip=None) -> _Sequence[_ods_ir.Value]:
  return CreateQubitsOp(result=result, qst_out=qst_out, amount=amount, qst_in=qst_in, loc=loc, ip=ip).results

@_ods_cext.register_operation(_Dialect)
class DeleteQubitsOp(_ods_ir.OpView):
  OPERATION_NAME = "jasp.delete_qubits"

  _ODS_REGIONS = (0, True)

  def __init__(self, out_qst, qubits, in_qst, *, loc=None, ip=None):
    operands = []
    results = []
    attributes = {}
    regions = None
    operands.append(qubits)
    operands.append(in_qst)
    _ods_context = _ods_get_default_loc_context(loc)
    results.append(out_qst)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def qubits(self):
    return self.operation.operands[0]

  @builtins.property
  def in_qst(self):
    return self.operation.operands[1]

  @builtins.property
  def out_qst(self):
    return self.operation.results[0]

def delete_qubits(out_qst, qubits, in_qst, *, loc=None, ip=None) -> _ods_ir.Value:
  return DeleteQubitsOp(out_qst=out_qst, qubits=qubits, in_qst=in_qst, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class FuseOp(_ods_ir.OpView):
  OPERATION_NAME = "jasp.fuse"

  _ODS_REGIONS = (0, True)

  def __init__(self, result, operand1, operand2, *, loc=None, ip=None):
    operands = []
    results = []
    attributes = {}
    regions = None
    operands.append(operand1)
    operands.append(operand2)
    _ods_context = _ods_get_default_loc_context(loc)
    results.append(result)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def operand1(self):
    return self.operation.operands[0]

  @builtins.property
  def operand2(self):
    return self.operation.operands[1]

  @builtins.property
  def result(self):
    return self.operation.results[0]

def fuse(result, operand1, operand2, *, loc=None, ip=None) -> _ods_ir.Value:
  return FuseOp(result=result, operand1=operand1, operand2=operand2, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class GetQubitOp(_ods_ir.OpView):
  OPERATION_NAME = "jasp.get_qubit"

  _ODS_REGIONS = (0, True)

  def __init__(self, result, qb_array, position, *, loc=None, ip=None):
    operands = []
    results = []
    attributes = {}
    regions = None
    operands.append(qb_array)
    operands.append(position)
    _ods_context = _ods_get_default_loc_context(loc)
    results.append(result)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def qb_array(self):
    return self.operation.operands[0]

  @builtins.property
  def position(self):
    return self.operation.operands[1]

  @builtins.property
  def result(self):
    return self.operation.results[0]

def get_qubit(result, qb_array, position, *, loc=None, ip=None) -> _ods_ir.Value:
  return GetQubitOp(result=result, qb_array=qb_array, position=position, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class GetSizeOp(_ods_ir.OpView):
  OPERATION_NAME = "jasp.get_size"

  _ODS_REGIONS = (0, True)

  def __init__(self, size, qb_array, *, loc=None, ip=None):
    operands = []
    results = []
    attributes = {}
    regions = None
    operands.append(qb_array)
    _ods_context = _ods_get_default_loc_context(loc)
    results.append(size)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def qb_array(self):
    return self.operation.operands[0]

  @builtins.property
  def size(self):
    return self.operation.results[0]

def get_size(size, qb_array, *, loc=None, ip=None) -> _ods_ir.Value:
  return GetSizeOp(size=size, qb_array=qb_array, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class MeasureOp(_ods_ir.OpView):
  OPERATION_NAME = "jasp.measure"

  _ODS_REGIONS = (0, True)

  def __init__(self, meas_res, out_qst, meas_q, in_qst, *, loc=None, ip=None):
    operands = []
    results = []
    attributes = {}
    regions = None
    operands.append(meas_q)
    operands.append(in_qst)
    _ods_context = _ods_get_default_loc_context(loc)
    results.append(meas_res)
    results.append(out_qst)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def meas_q(self):
    return self.operation.operands[0]

  @builtins.property
  def in_qst(self):
    return self.operation.operands[1]

  @builtins.property
  def meas_res(self):
    return self.operation.results[0]

  @builtins.property
  def out_qst(self):
    return self.operation.results[1]

def measure(meas_res, out_qst, meas_q, in_qst, *, loc=None, ip=None) -> _Sequence[_ods_ir.Value]:
  return MeasureOp(meas_res=meas_res, out_qst=out_qst, meas_q=meas_q, in_qst=in_qst, loc=loc, ip=ip).results

@_ods_cext.register_operation(_Dialect)
class QuantumGateOp(_ods_ir.OpView):
  OPERATION_NAME = "jasp.quantum_gate"

  _ODS_REGIONS = (0, True)

  def __init__(self, out_qst, gate_type, gate_operands, in_qst, *, loc=None, ip=None):
    operands = []
    results = []
    attributes = {}
    regions = None
    operands.extend(_get_op_results_or_values(gate_operands))
    operands.append(in_qst)
    _ods_context = _ods_get_default_loc_context(loc)
    attributes["gate_type"] = (gate_type if (
    isinstance(gate_type, _ods_ir.Attribute) or
    not _ods_ir.AttrBuilder.contains('StrAttr')) else
      _ods_ir.AttrBuilder.get('StrAttr')(gate_type, context=_ods_context))
    results.append(out_qst)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def gate_operands(self):
    _ods_variadic_group_length = len(self.operation.operands) - 2 + 1
    return self.operation.operands[0:0 + _ods_variadic_group_length]

  @builtins.property
  def in_qst(self):
    _ods_variadic_group_length = len(self.operation.operands) - 2 + 1
    return self.operation.operands[1 + _ods_variadic_group_length - 1]

  @builtins.property
  def gate_type(self):
    return self.operation.attributes["gate_type"]

  @gate_type.setter
  def gate_type(self, value):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["gate_type"] = value

  @builtins.property
  def out_qst(self):
    return self.operation.results[0]

def quantum_gate(out_qst, gate_type, gate_operands, in_qst, *, loc=None, ip=None) -> _ods_ir.Value:
  return QuantumGateOp(out_qst=out_qst, gate_type=gate_type, gate_operands=gate_operands, in_qst=in_qst, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class ResetOp(_ods_ir.OpView):
  OPERATION_NAME = "jasp.reset"

  _ODS_REGIONS = (0, True)

  def __init__(self, out_qst, qubits, in_qst, *, loc=None, ip=None):
    operands = []
    results = []
    attributes = {}
    regions = None
    operands.append(qubits)
    operands.append(in_qst)
    _ods_context = _ods_get_default_loc_context(loc)
    results.append(out_qst)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def qubits(self):
    return self.operation.operands[0]

  @builtins.property
  def in_qst(self):
    return self.operation.operands[1]

  @builtins.property
  def out_qst(self):
    return self.operation.results[0]

def reset(out_qst, qubits, in_qst, *, loc=None, ip=None) -> _ods_ir.Value:
  return ResetOp(out_qst=out_qst, qubits=qubits, in_qst=in_qst, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class SliceOp(_ods_ir.OpView):
  OPERATION_NAME = "jasp.slice"

  _ODS_REGIONS = (0, True)

  def __init__(self, result, qb_array, start, end, *, loc=None, ip=None):
    operands = []
    results = []
    attributes = {}
    regions = None
    operands.append(qb_array)
    operands.append(start)
    operands.append(end)
    _ods_context = _ods_get_default_loc_context(loc)
    results.append(result)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def qb_array(self):
    return self.operation.operands[0]

  @builtins.property
  def start(self):
    return self.operation.operands[1]

  @builtins.property
  def end(self):
    return self.operation.operands[2]

  @builtins.property
  def result(self):
    return self.operation.results[0]

def slice(result, qb_array, start, end, *, loc=None, ip=None) -> _ods_ir.Value:
  return SliceOp(result=result, qb_array=qb_array, start=start, end=end, loc=loc, ip=ip).result
